import { Logger } from './utils/logger';
import {
    ClientMCPConfig,
    ChatChunk,
    ConversationState
} from './types';
import OpenAI from "openai";
import {
    Tool as AnthropicTool
} from "@anthropic-ai/sdk/resources/messages/messages.mjs";
import { Client } from "@modelcontextprotocol/sdk/client/index.js";
import { StdioClientTransport } from "@modelcontextprotocol/sdk/client/stdio.js";
import { SSEClientTransport } from "@modelcontextprotocol/sdk/client/sse.js";
import { convertContext } from "./utils/convertContext";

/**
 * Production-ready MCP (Model Context Protocol) client for AI conversations with tool support
 * 
 * @example
 * ```typescript
 * const client = new ClientMCP({
 *   apiKey: "your-api-key",
 *   baseUrl: "https://your-model-context-protocol-server.com",
 *   model: "gemini-2.0-flash",
 *   debug: true
 * });
 * 
 * await client.connectToServer("http://localhost:3003/mcp");
 * 
 * for await (const chunk of client.chat("Hello, how are you?")) {
 *   console.log(chunk.choices[0]?.delta?.content);
 * }
 * ```
 */
export class ClientMCP {
    private static readonly DEFAULT_TIMEOUT = 30000;
    private static readonly DEFAULT_MODEL = "gemini-2.0-flash";
    private static readonly MAX_RECURSION_DEPTH = 5;

    private readonly config: Required<Pick<ClientMCPConfig, 'timeout' | 'debug'>> & {
        apiKey: string;
        model: string;
        baseUrl?: string;
    };

    private readonly logger: Logger;
    private readonly ai: OpenAI;
    private readonly mcp: Client;

    private transport: StdioClientTransport | SSEClientTransport | null = null;
    private tools: AnthropicTool[] = [];
    private functionDeclarations: OpenAI.Chat.Completions.ChatCompletionTool[] = [];
    private messages: OpenAI.Chat.Completions.ChatCompletionMessageParam[] = [];
    private abortController: AbortController;
    private isConnected = false;
    private recentConversationLength = 0;

    constructor(config: ClientMCPConfig) {
        this.validateConfig(config);
        this.config = this.normalizeConfig(config);
        this.logger = Logger.getInstance(this.config.debug);
        this.ai = this.initializeOpenAI();
        this.mcp = this.initializeMCP();
        this.abortController = new AbortController();
        this.addMessage(config?.systemMessages || "You are a smart AI assistant. You have access to tools to help you with your tasks. so you have to plan your actions carefully that you can use those tools and complete your task. \n" +
            "You have to refactor your plan if it is not working.\n" +
            "You have to create plan at the start of any task \n" +
            "You can use multiple tools to complete your task. \n" +
            "You can use tools one after one, or simultaneously. \n",
            "system");

    }

    /**
     * Establishes connection to MCP server
     * 
     * @param serverPath - URL or string path to the MCP server
     * @param sessionId - Optional session ID for reconnection
     */
    public async connectToServer(serverPath: string | URL, sessionId?: string): Promise<void> {
        if (this.isConnected) {
            this.logger.warn('Already connected to MCP server');
            return;
        }

        try {
            this.transport = await this.createTransport(serverPath, sessionId);
            await this.mcp.connect(this.transport);
            await this.loadTools();
            this.isConnected = true;
            this.logger.info(`Successfully connected to MCP server: ${serverPath}`);
        } catch (error) {
            this.logger.error('Failed to connect to MCP server:', error);
            throw new Error(`MCP connection failed: ${error instanceof Error ? error.message : 'Unknown error'}`);
        }
    }

    /**
     * Disconnects from MCP server
     */
    public async disconnect(): Promise<void> {
        if (!this.isConnected) return;

        try {
            if (this.transport) {
                await this.transport.close();
                this.transport = null;
            }
            this.isConnected = false;
            this.logger.info('Disconnected from MCP server');
        } catch (error) {
            this.logger.error('Error during disconnect:', error);
        }
    }

    /**
     * Performs streaming chat completion
     * 
     * @param options - Options for the chat completion
     * @returns Async generator of chat completion chunks
     */
    public async *chatCompletionStream(options: {
        messages: OpenAI.Chat.Completions.ChatCompletionMessageParam[];
        reasoningEffort?: "low" | "medium" | "high";
        tools?: OpenAI.Chat.Completions.ChatCompletionTool[];
        toolChoice?: OpenAI.Chat.Completions.ChatCompletionCreateParams['tool_choice'];
    }): AsyncGenerator<OpenAI.Chat.Completions.ChatCompletionChunk> {
        try {
            const stream = await this.ai.chat.completions.create({
                model: this.config.model,
                messages: options.messages,
                ...(options.reasoningEffort && { reasoning_effort: options.reasoningEffort }),
                ...(options.tools && { tools: options.tools }),
                ...(options.toolChoice && { tool_choice: options.toolChoice }),
                stream: true,
            });

            for await (const chunk of stream) {
                yield chunk;
            }
        } catch (error) {
            this.logger.error('Streaming chat completion failed:', error);
            throw this.handleApiError(error);
        }
    }

    /**
     * Main chat interface with automatic tool handling and conversation management
     * 
     * @param content - User input message
     * @param options - Options for the chat with maxDepth and autoSummarize
     * Where maxDepth is the maximum depth of recursion and autoSummarize is whether to automatically summarize the conversation
     * @returns Async generator of chat completion chunks and tool calls
     */
    public async *chat(
        content: string,
        options: {
            maxDepth?: number;
            autoSummarize?: boolean;
        } = {}
    ): AsyncGenerator<OpenAI.Chat.Completions.ChatCompletionChunk | ChatChunk> {
        const { maxDepth = ClientMCP.MAX_RECURSION_DEPTH, autoSummarize = false } = options;
        this.recentConversationLength = 0;
        yield* this.chatInternal(content, 0, maxDepth, autoSummarize);
    }

    /**
     * Cancels all ongoing requests
     */
    public cancelRequests(): void {
        this.abortController.abort();
        this.abortController = new AbortController();
        this.logger.info('All requests cancelled');
    }

    /**
     * Gets current conversation state
     * 
     * @returns Current conversation state
     */
    public getConversationState(): ConversationState {
        return {
            messageCount: this.messages.length,
            toolsAvailable: this.tools.length,
            isConnected: this.isConnected,
            lastActivity: new Date()
        };
    }

    /**
     * Clears conversation history
     */
    public clearConversation(): void {
        const systemMessages = this.messages.filter(msg => msg.role === 'system');
        this.messages = systemMessages;
        this.logger.info('Conversation history cleared');
    }

    // Private methods

    private validateConfig(config: ClientMCPConfig): void {
        if (!config.apiKey?.trim()) {
            throw new Error('API key is required and cannot be empty');
        }
    }

    private normalizeConfig(config: ClientMCPConfig) {
        return {
            timeout: config.timeout ?? ClientMCP.DEFAULT_TIMEOUT,
            debug: config.debug ?? false,
            apiKey: config.apiKey.trim(),
            model: config.model ?? ClientMCP.DEFAULT_MODEL,
            baseUrl: config.baseUrl
        };
    }

    private initializeOpenAI(): OpenAI {
        return new OpenAI({
            apiKey: this.config.apiKey,
            ...(this.config.baseUrl && { baseURL: this.config.baseUrl }),
            timeout: this.config.timeout
        });
    }

    private initializeMCP(): Client {
        return new Client(
            { name: "mcp-client-cli", version: "1.0.0" },
            { capabilities: { tools: {} } }
        );
    }

    private async createTransport(
        serverPath: string | URL,
        sessionId?: string
    ): Promise<StdioClientTransport | SSEClientTransport> {
        if (typeof serverPath === 'string') {
            return this.createStdioTransport(serverPath);
        } else {
            return this.createSSETransport(serverPath, sessionId);
        }
    }

    private createStdioTransport(serverPath: string): StdioClientTransport {
        const isJavaScript = serverPath.endsWith('.js');
        const isPython = serverPath.endsWith('.py');

        if (!isJavaScript && !isPython) {
            throw new Error('Server script must be a .js or .py file');
        }

        const command = isPython
            ? (process.platform === "win32" ? "python" : "python3")
            : process.execPath;

        return new StdioClientTransport({
            command,
            args: [serverPath]
        });
    }

    private createSSETransport(serverUrl: URL, sessionId?: string): SSEClientTransport {
        return new SSEClientTransport(serverUrl, {
            requestInit: {
                headers: {
                    'Accept': 'text/event-stream',
                    'Content-Type': 'application/json',
                    ...(sessionId && { 'mcp-session-id': sessionId })
                }
            }
        });
    }


    public async loadLocalFunctionDeclarations(): Promise<void> {
        const generate_custom_tool = {
            "type": "function",
            "function": {
                "name": "generate_custom_tool",
                "description": "You will create a python function based on the user's request. \n" +
                "The function will be used to complete the user's request. \n" +
                "The function will be added to the tools list. \n" +
                "The function will be used in the next step." +
                "This tool is only be used when there is no tool available to do the specified task." + 
                "If you do not have the functionality for the user's request, you can use this tool to create a python function to do the specified task.\n",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "tool_name": { 
                            "type": "string",
                            "description": "The name of the tool"
                        },
                        "tool_description": { 
                            "type": "string",
                            "description": "The description of the tool"
                        },
                        "tool_input_schema": {
                            "type": "object",
                            "description": "The input schema of the tool"
                        },
                        "tool_python_code": { 
                            "type": "string",
                            "description": "The python code of the tool"
                        }
                    },
                    "required": ["tool_name", "tool_description", "tool_input_schema", "tool_python_code"]
                }
            }
        } as OpenAI.Chat.Completions.ChatCompletionTool;
        
        this.functionDeclarations.push(generate_custom_tool);
    }

    private async loadTools(): Promise<void> {
        const toolsResult = await this.mcp.listTools();

        this.tools = toolsResult.tools.map(tool => ({
            name: tool.name,
            description: tool.description,
            input_schema: tool.inputSchema
        }));

        this.functionDeclarations = convertContext(this.tools);

        // this.loadLocalFunctionDeclarations();

        this.logger.info(`Loaded ${this.tools.length} tools:`,
            this.tools.map(tool => `${tool.name}: ${tool.description}`).join(', ')
        );
    }


    private async *chatInternal(
        content: string,
        depth: number,
        maxDepth: number,
        autoSummarize: boolean
    ): AsyncGenerator<OpenAI.Chat.Completions.ChatCompletionChunk | ChatChunk> {
        if (depth >= maxDepth) {
            yield* this.handleMaxDepthReached(content, maxDepth);
            return;
        }

        this.addMessage(content, "user");
        let assistantResponse = '';

        // Stream the main conversation
        for await (const chunk of this.chatCompletionStream({
            messages: this.messages,
            tools: this.functionDeclarations.length > 0 ? this.functionDeclarations : undefined
        })) {
            const textContent = chunk.choices[0]?.delta?.content || '';
            if (textContent) {
                assistantResponse += textContent;
                yield chunk;
            }

            // Handle tool calls
            const toolCalls = chunk.choices[0]?.delta?.tool_calls;
            if (toolCalls?.length) {
                for (const toolCall of toolCalls) {
                    yield {
                        choices: [{
                            delta: {
                                content: JSON.stringify({
                                    tool: toolCall.function?.name,
                                    args: JSON.parse(toolCall.function?.arguments || "{}"),
                                    executed: false
                                }, null, 2),
                                tool_calls: [{
                                    function: {
                                        name: toolCall.function?.name!,
                                        arguments: toolCall.function?.arguments!
                                    }
                                }]
                            }
                        }]
                    }
                }
                const toolResults = await this.executeToolCalls(toolCalls);
                for (const result of toolResults) {
                    autoSummarize = true;
                    if (result) yield result;
                }
            }
        }

        if (assistantResponse.trim()) {
            this.addMessage(assistantResponse.trim(), 'assistant');
        }

        // Check if continuation is needed
        const shouldContinue = await this.shouldContinueConversation();
        if (shouldContinue.continue && shouldContinue.nextMessage) {
            yield* this.chatInternal(shouldContinue.nextMessage, depth + 1, maxDepth, autoSummarize);
        } else if (autoSummarize) {
            yield* this.generateConversationSummary();
        }
    }

    private async *handleMaxDepthReached(
        originalContent: string,
        maxDepth: number
    ): AsyncGenerator<OpenAI.Chat.Completions.ChatCompletionChunk> {
        this.logger.warn(`Max depth (${maxDepth}) reached for conversation`);

        const contextMessage = `Maximum conversation depth reached. Original request: "${originalContent}". ` +
            `Please provide a final response based on the conversation context or indicate if additional input is needed.`;

        yield* this.chatCompletionStream({
            messages: [...this.messages, { role: "user", content: contextMessage }]
        });
    }

    private async executeToolCalls(
        toolCalls: OpenAI.Chat.Completions.ChatCompletionChunk.Choice.Delta.ToolCall[]
    ): Promise<(ChatChunk | null)[]> {
        const toolPromises = toolCalls
            .filter(call => call.function?.name && call.function?.arguments)
            .map(call => this.executeToolCall(call));

        return Promise.all(toolPromises);
    }

    private async executeToolCall(
        toolCall: OpenAI.Chat.Completions.ChatCompletionChunk.Choice.Delta.ToolCall
    ): Promise<ChatChunk | null> {
        const { name, arguments: args } = toolCall.function!;

        if(name === "generate_custom_tool"){
            const toolArgs = JSON.parse(args || "{}");
            const result = await this.generateCustomTool(toolArgs);
            return {
                choices: [{
                    delta: {
                        content: result,
                        tool_calls: [{ function: { name: name!, arguments: args || "{}" } }]
                    }
                }]
            };
        }

        try {
            const toolArgs = JSON.parse(args || "{}");
            const result = await this.mcp.callTool({ name: name!, arguments: toolArgs });

            const resultContent = this.formatToolResult(name!, toolArgs, result.content);
            this.addMessage(resultContent, 'assistant');

            return {
                choices: [{
                    delta: {
                        content: resultContent,
                        tool_calls: [{ function: { name: name!, arguments: args || "{}" } }]
                    }
                }]
            };
        } catch (error) {
            this.logger.error(`Tool execution failed for ${name}:`, error);
            const errorContent = `Tool "${name}" execution failed: ${error instanceof Error ? error.message : 'Unknown error'}`;

            return {
                choices: [{
                    delta: {
                        content: errorContent,
                        tool_calls: [{ function: { name: name!, arguments: args || "{}" } }]
                    }
                }]
            };
        }
    }

    private formatToolResult(name: string, args: Record<string, unknown>, result: unknown): string {
        const resJson = {
            tool: name,
            args,
            result,
            executed: true
        }
        return JSON.stringify(resJson, null, 2);
    }

    private async shouldContinueConversation(): Promise<{ continue: boolean; nextMessage?: string }> {
        try {
            const tools: OpenAI.Chat.Completions.ChatCompletionTool[] = [{
                "type": "function",
                "function": {
                    "name": "evaluate_conversation_status",
                    "description": "Analyze the current conversation state. \n" +
                        "Determine if the user's original request has been fully addressed or if additional steps are needed. \n" +
                        "If more work is required, suggest the next logical step. \n" +
                        "If there any task/steps remains, suggest the next logical step. \n" +
                        "Those tasks/steps can be completed by those tools in the next step but first you have to return 'continue': true and 'nextMessage': 'the next message to process if continuation is needed' \n" +
                        "The tools are: \n" +
                        this.tools.map((tool, index) => `${index + 1}. ${tool.name}: ${tool.description}`).join('\n'),
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "continue": {
                                "type": "boolean",
                                "description": "Whether the conversation should continue"
                            },
                            "nextMessage": {
                                "type": "string",
                                "description": "The next message to process if continuation is needed"
                            }
                        },
                        "required": ["continue"]
                    }
                }
            }];

            const evaluationPrompt = "Have you completed the user's original request? If not, suggest the next logical step.";

            let functionArgs = '';
            for await (const chunk of this.chatCompletionStream({
                messages: [...this.messages, { role: "user", content: evaluationPrompt }],
                tools
            })) {
                const toolCall = chunk.choices[0]?.delta?.tool_calls?.[0];
                if (toolCall?.function?.name === "evaluate_conversation_status") {
                    functionArgs += toolCall.function.arguments || '';
                }
            }

            const result = JSON.parse(functionArgs || '{"continue": false}');
            return {
                continue: result.continue || false,
                nextMessage: result.nextMessage
            };
        } catch (error) {
            this.logger.error('Error evaluating conversation status:', error);
            return { continue: false };
        }
    }

    private async *generateConversationSummary(): AsyncGenerator<OpenAI.Chat.Completions.ChatCompletionChunk> {
        const summaryPrompt = "Please provide a concise summary of our conversation, " +
            "highlighting the key points, decisions made, and any important outcomes." +
            `Do not use summary word in the response.  ***Just provide the summary for last ${this.recentConversationLength + 1} messages. ***`;
        this.logger.debug("messages::", [...this.messages, { role: "user", content: summaryPrompt }]);
        yield* this.chatCompletionStream({
            messages: [...this.messages, { role: "user", content: summaryPrompt }]
        });
    }

    private addMessage(content: string, role: "user" | "assistant" | "system"): void {
        this.messages.push({ role, content });
        this.recentConversationLength = this.recentConversationLength + 1;
    }

    private handleApiError(error: unknown): Error {
        if (error instanceof Error) {
            if (error.name === 'AbortError') {
                return new Error('Request was cancelled');
            }
            return error;
        }
        return new Error('Unknown API error occurred');
    }


    private async generateCustomTool(args: {
        tool_name: string;
        tool_description: string;
        tool_input_schema: Record<string, unknown>;
        tool_python_code: string;
    }): Promise<string> {
        // const tool = {
        //     name: args.tool_name,
        //     description: args.tool_description,
        //     input_schema: args.tool_input_schema
        // };
        
        this.functionDeclarations.push({
            "type": "function",
            "function": {
                "name": args.tool_name,
                "description": args.tool_description,
                "parameters": args.tool_input_schema
            }
        });
        return args.tool_python_code;
    }
}